/* Copyright 2019 The Matrix.org Foundation C.I.C
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* delete room keys that belong to deleted room key version, or to room key
 * versions that don't exist (anymore)
 */
DELETE FROM e2e_room_keys
WHERE version NOT IN (
  SELECT version
  FROM e2e_room_keys_versions
  WHERE e2e_room_keys.user_id = e2e_room_keys_versions.user_id
  AND e2e_room_keys_versions.deleted = 0
);
